import json
import logging
import os
import pathlib
import stat
import subprocess
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Union

import requests
from messaging.router import Router, on_message
from pycapo import CapoConfig

from workspaces.system.schema import AbstractFile
from workspaces.workflow.enum import WorkflowRequestState
from workspaces.workflow.schema import WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.schema_interfaces import WorkflowIF, WorkflowRequestIF
from workspaces.workflow.services.interfaces import WorkflowInfoIF, WorkflowServiceIF

logger = logging.getLogger(__name__)


class WorkflowServiceRESTClient(WorkflowServiceIF):
    def __init__(self):
        self.url = CapoConfig().settings("edu.nrao.archive.workspaces.WorkflowSettings").serviceUrl

    def execute(self, request: WorkflowRequestIF):
        result = requests.post(
            f"{self.url}/workflows/{request.workflow_name}"
            f"/requests/{request.workflow_request_id}/submit"
        )
        logger.info(
            "Got result %s with type %s and body %s",
            result,
            type(result),
            result.content,
        )
        return result.json()

    def attach_file_to_request(self, request: WorkflowRequestIF, file: AbstractFile):
        requests.post(
            f"{self.url}/workflows/{request.workflow_name}"
            f"/requests/{request.workflow_request_id}"
            f"/files/{file.filename}",
            data=file.content,
        )

    def create_workflow_request(
        self, workflow: Union[str, WorkflowIF], argument: Dict
    ) -> WorkflowRequestIF:
        # 1. Handle the type ambiguity with the workflow argument
        workflow_name = workflow if type(workflow) == type(str()) else workflow.workflow_name

        # 2. Make the request
        result = requests.post(
            f"{self.url}/workflows/{workflow_name}/requests/create",
            json=argument,
        ).json()

        return WorkflowRequest.from_json(result)


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: WorkflowInfoIF):
        self.message_router = Router("workflow")
        self.message_router.register(self)
        self.info = info
        self.useCondor = CapoConfig().getboolean(
            "edu.nrao.archive.workspaces.ProcessingSettings.useCondor"
        )
        self.root_directory = (
            CapoConfig().settings("edu.nrao.archive.workspaces.ProcessingSettings").rootDirectory
        )

    def attach_file_to_request(self, request: WorkflowRequestIF, file: AbstractFile):
        self.info.save_file(request=request, filename=file.filename, content=file.content)

    def create_workflow_request(self, workflow, argument) -> WorkflowRequestIF:
        # return self.info.save_request(WorkflowRequest(workflow_name=workflow, argument=argument))
        return self.info.create_workflow_request(workflow, argument)

    def execute(self, request: WorkflowRequest):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :param files: additional files to execute with request
        :return:
        """
        # 1. look up workflow, returns workflow
        logger.info("Looking up workflow %s", request.workflow_name)
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # 1. create a temporary directory, needs to exist before template rendering
        temp_folder = self._make_temp_directory()

        # 2. render templates to files, returns list of rendered files
        logger.info("Rendering files")
        logger.info("Type of request.argument is %s", type(request.argument))

        # Aggrandize the argument with the workflow request ID
        request.argument["request_id"] = request.workflow_request_id

        # render all the templates
        if request.argument["need_project_metadata"] is True:
            wf_metadata = subprocess.Popen(
                ["aat_wrest", "-sc", request.argument["product_locator"]], stdout=subprocess.PIPE
            )
            wf_json = wf_metadata.communicate()[0]
            merged_args = {
                **json.loads(wf_json),
                **request.argument,
                "relative_path": str(temp_folder).split("/")[-1],
                "root_directory": self.root_directory,
            }
            templated_files = definition.render_templates(merged_args)
        else:
            templated_files = definition.render_templates(request.argument)

        for file in templated_files:
            self.info.save_file(request=request, filename=file.filename, content=file.content)

        # 3. Combine the templated with the request files
        workflow_files = templated_files + request.files

        # 4. prepare files for condor execution
        if not request.results_dir:
            # Create temp results directory if the request doesn't already have one (from a previous execution attempt)
            request.results_dir = str(
                self._prepare_files_for_condor(workflow_files, temp_folder).absolute()
            )
            self.info.save_request(request)

        # 5. execute condor and retrieve log file
        log_file = self._execute_prepared(Path(request.results_dir))

        # 6. start reading the log file (background it)
        subprocess.run(
            ["wf_monitor", log_file, str(request.workflow_request_id)],
            cwd=request.results_dir,
        )

        return request

    def _make_temp_directory(self) -> Path:
        """
        Create tmp folder for workflow

        :return: path to tmp folder
        """
        temp_folder_path = self.root_directory
        if not os.path.isdir(temp_folder_path):
            os.mkdir(temp_folder_path)
        temp_folder = Path(mkdtemp(dir=temp_folder_path))
        logger.info("Settled on temp folder %s", temp_folder)
        return temp_folder

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile], temp_folder: Path) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. spool each of the temp files to tmp directory
        for file in files:
            path = temp_folder / file.filename
            logger.info("Writing %s", path)
            path.write_bytes(file.content)

        # 2. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            logger.info("Making %s executable", file)
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    def _execute_prepared(self, folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        logger.info("executing on folder %s", folder)

        # some file in here should end in .dag; that file is our dagman input
        # TODO: not currently using DAG files for vulture
        # TODO: 3.25.21 - vulture reports malformed file for null.dag. Will need checking
        # dagman = list(folder.glob("*.dag"))[0]
        # logger.info("dagman file %s exists.", dagman)

        # need .condor file for vulture submission
        condor = list(folder.glob("*.condor"))[0]
        logger.info("condor file %s exists.", condor)

        # ensure the log file exists
        logfile = self._get_logfile_name(condor)
        logger.info("log file %s exists.", logfile)
        logfile.touch()

        # submit
        if self.useCondor:
            logger.info("submitting to condor...")
            subprocess.run(
                ["condor_submit", str(condor)],
                cwd=str(folder.absolute()),
                preexec_fn=self._switch_to_submituser,
            )
        else:
            # vulture is a workaround for testing locally without submitting to condor
            logger.info("submitting to vulture...")
            subprocess.run(["vulture", "job", "execute", str(condor)], cwd=str(folder.absolute()))

        # return the logfile
        return logfile

    @staticmethod
    def _switch_to_submituser():
        """
        Helper function that is meant to be called as a `preexec_fn` for `subprocess.run`; sets gid and uid
        of subprocess to those of the submituser so that jobs can be properly submitted
        """
        submituser_gid = 6000
        submituser_uid = 6000

        os.setgid(submituser_gid)
        os.setuid(submituser_uid)

    @staticmethod
    def _get_logfile_name(jobfile_name: Path) -> Path:
        """
        Read HTCondor job file and get the log file name, if it exists

        :param jobfile_name: Job file name
        :return: Log file name or "{job-name}.log"
        """
        with open(jobfile_name, "rt") as f:
            for line in f:
                if line.startswith("log = "):
                    _, logfile_name = line.split(" = ")
                    return Path(logfile_name.strip())
        return Path(f"{jobfile_name.stem}.log")

    @on_message(service="workflow", type="workflow-complete")
    def propagate_delivery(self, **message: Dict):
        # look up the workflow request to get the path
        request = self.info.lookup_workflow_request(message["subject"]["workflow_request_id"])

        # find the path to the delivery.json file
        delivery_file = pathlib.Path(request.results_dir) / "delivery.json"

        # does it exist? if so, we need to send out a new event with its contents
        if delivery_file.exists():
            logger.info("Found delivery.json file, sending delivery notification")
            with open(delivery_file, "r") as delivery_fo:
                delivery_info = json.load(delivery_fo)

                # construct the message
                message = {
                    "subject": message["subject"],
                    "service": "workflow",
                    "type": "delivery",
                    "delivery": delivery_info,
                }

                # send the message
                self.message_router.send_message(**message)
        else:
            logger.info("No delivery.json found")

    @on_message(service="workflow")
    def on_workflow_event(self, **message: Dict):
        logger.info(f"RECEIVED WORKFLOW MESSAGE: {message}")
        msg_subject = message["subject"]
        if msg_subject["type"] == "WorkflowRequest":
            request_id = msg_subject["workflow_request_id"]
            request = self.info.lookup_workflow_request(request_id)

            # 2. update request record with new status
            if message["type"] == "workflow-executing":
                status = WorkflowRequestState.Running.name
            elif message["type"] == "workflow-complete":
                status = WorkflowRequestState.Complete.name
            elif message["type"] == "workflow-failed":
                status = WorkflowRequestState.Failed.name
            else:
                status = "Unknown"
                # status = WorkflowRequestState.Running.name

            logger.info(
                "Updating state on workflow request %s to %s...",
                request.workflow_request_id,
                status,
            )
            request.update_status(status)
        else:
            logger.warning(f"Message {message} does not concern a workflow request. Ignoring.")
